/*
        FILE :                  3dstruct.h

        DESCRIPTION :           Basic 3d structures

        AUTHOR :                Trax / Aspirine (c), 1999

        VERSION :               0.1b

*/

#ifndef _3DSTRUCT_H_INCLUDED
#define _3DSTRUCT_H_INCLUDED

//// DEFINES //////////////////////////////////////////////////////////////////

#define MAXTEXTURES     8

//// STRUCTS //////////////////////////////////////////////////////////////////

// VERTEX TYPE //
typedef struct
{
        float x, y, z;
}VERTEX;

///////////////////////////////////////////////////////////////////////////////

// MATRIX TYPE //
typedef float MATRIX[3][3];

///////////////////////////////////////////////////////////////////////////////

// TEXTURE COORDS TYPE //
typedef struct
{
        float u, v;     // texture u, texture v
}UV;

///////////////////////////////////////////////////////////////////////////////

// FACE TYPE //
typedef struct
{
        int a, b, c;
        VERTEX facenormal;
        unsigned int flags;
}FACE;

/*
        values of flags:

        switch (flags & 15)
        {
                case 0 : flat
                case 1 : gouraud
                case 2 : texture
                case 3 : gouraud+texture
                case 4 : phong
                case 5 : envmap
                default: reserved for future expansions
        }
*/

///////////////////////////////////////////////////////////////////////////////

// LIGHT TYPE //
typedef struct
{
        VERTEX direction; 
}LIGHT;

///////////////////////////////////////////////////////////////////////////////

// POLYGON VERTEX TYPE
typedef struct
{
        int x, y, z, u, v, c, r, g, b;  // r, g , b = c(olor)
        float nx, ny, nz;               // normal x, normal y, normal z
}POL_VERTEX;

///////////////////////////////////////////////////////////////////////////////

// IMAGE TYPE //
typedef struct
{
        char *filename;
        unsigned char *palette;
        unsigned char *data;
        unsigned short xsize, ysize;
}IMAGE;

///////////////////////////////////////////////////////////////////////////////

// ORDER TYPE -> ZSORT
typedef struct
{
        int order, face; 
        unsigned short object; 
}ORDER_STRUCT;

///////////////////////////////////////////////////////////////////////////////

// CAMERA TYPE //
typedef struct
{
        float xpos, ypos, zpos, perspective, center_x, center_y, x_suhde, y_suhde;
        int screen_left,screen_right,screen_up,screen_down;
        MATRIX cam;
}CAMERA;

///////////////////////////////////////////////////////////////////////////////

// OBJECT TYPE //
typedef struct
{
        VERTEX *vertex;
        VERTEX *normal;
        UV *uv;
        FACE *face;
        unsigned int vertices, faces;
}OBJECT;

///////////////////////////////////////////////////////////////////////////////

// OBJECT LINK //
typedef struct
{
        OBJECT object;  /* vertices, faces, normals */
        unsigned int vertex_index; /* index to rot */
        unsigned int face_index; /* index to order */
        float xpos, ypos, zpos; /* object's position */
        MATRIX obj_matrix;
        VERTEX origin; /* object's position presented as a vertex */
        int hidden; // 1 yes, 0 no
}OBJECTLINK;

///////////////////////////////////////////////////////////////////////////////

// SCENE TYPE //
typedef struct
{
        OBJECTLINK *obut;   
        LIGHT light;  // light
        VERTEX *rot; // rotation
        VERTEX *nor; // rotation normals
        unsigned int total_objects;      
        unsigned int faces_in_scene;     
        unsigned int vertices_in_scene;  
        IMAGE textm[MAXTEXTURES];         
        int perspective;
        unsigned int total_rot_vertices;
        unsigned int total_faces;
}SCENE;




#endif
