/*
        FILE :                  matrix.cpp

        DESCRIPTION :           Implementation of matrix routines

        AUTHOR :                Trax / Aspirine (c), 1999

        VERSION :               0.1b
*/

//// INCLUDES /////////////////////////////////////////////////////////////////

#include "matrix.h"

//// FUNCTIONS ////////////////////////////////////////////////////////////////

/* 
    a b c     k l m     ak+bn+cq al+bo+cr am+bp+cs 
    d e f  *  n o p  =  dk+en+fq dl+eo+fr dm+ep+fs 
    h i j     q r s     hk+in+jq hl+io+jr hm+ip+js 
*/

void matrix_mul(MATRIX a, MATRIX b) /* return to a */
{
        MATRIX tmp;

        int i, j;
        for (i = 0; i < 3; i++)
                for (j = 0; j < 3; j++)
                        tmp[i][j] = a[i][0] * b[0][j] +
                                    a[i][1] * b[1][j] +
                                    a[i][2] * b[2][j];
        for (i = 0; i < 3; i++)
                for (j = 0; j < 3; j++)
                        a[i][j] = tmp[i][j];
}

///////////////////////////////////////////////////////////////////////////////

/*
                 a b c 0 
   (Xi+Yj+Zk) *  e f g 0  = (aX+eY+iZ+m)i + (bX+fY+jZ+n)j +
                 i j k 0    (cX+gY+kZ+o)k
                 m n o 1 
*/

/*inline*/ void matrix_mul_vector(VERTEX vector, VERTEX *result, MATRIX mat)
{
        VERTEX vec = vector;

        result->x = mat[0][0] * vec.x + mat[1][0] * vec.y + mat[2][0] * vec.z;
        result->y = mat[0][1] * vec.x + mat[1][1] * vec.y + mat[2][1] * vec.z;
        result->z = mat[0][2] * vec.x + mat[1][2] * vec.y + mat[2][2] * vec.z;
}

///////////////////////////////////////////////////////////////////////////////

/*
                      cy*cz          cy*sz          -sy    0 
                      sx*sy*cz-cx*sz sx*sy*sz+cx*cz  sx*cy 0 
       [X]*[Y]*[Z] =  cx*sy*cz+sx*sz cx*sy*sz-sx*cz  cx*cy 0 
                      0              0               0     1 
*/

void matrix_update(MATRIX m, float sx, float sy, float sz,
                   float cx, float cy, float cz)
{
        m[0][0] = cy * cz;
        m[0][1] = cy * sz;
        m[0][2] = -sy;

        m[1][0] = sx * sy * cz - cx * sz;
        m[1][1] = sx * sy * sz + cx * cz;
        m[1][2] = sx * cy;

        m[2][0] = cx * sy * cz + sx * sz;
        m[2][1] = cx * sy * sz - sx * cz;
        m[2][2] = cx * cy;
}

///////////////////////////////////////////////////////////////////////////////

/*
      1 0 0 0 
      0 1 0 0 
      0 0 1 0 
      0 0 0 1 
*/

void matrix_normalize(MATRIX m)
{
        m[0][0] = 1; m[1][0] = 0; m[2][0] = 0;
        m[0][1] = 0; m[1][1] = 1; m[2][1] = 0;
        m[0][2] = 0; m[1][2] = 0; m[2][2] = 1;
}

/*//////////////////////////////////////////////////////////////////////////*/

