/*
        FILE :                  object.cpp

        DESCRIPTION :           Implementation of objects routines

        AUTHOR :                Trax / Aspirine (c), 1999

        VERSION :               0.1b
*/

//// INCLUDES /////////////////////////////////////////////////////////////////

#include "object.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

//// FUNCTIONS ////////////////////////////////////////////////////////////////

void dispose_object(SCENE *scene, int nro)
{
        if (scene->obut[nro].object.vertex != NULL)
                free(scene->obut[nro].object.vertex);

        if (scene->obut[nro].object.face != NULL)
                free(scene->obut[nro].object.face);

        if (scene->obut[nro].object.normal != NULL)
                free(scene->obut[nro].object.normal);

        if (scene->obut[nro].object.uv != NULL)
                free(scene->obut[nro].object.uv);

        memset(&scene->obut[nro].object, 0, sizeof(OBJECT));

        while ( (scene->total_objects > 0) &&
                (scene->obut[scene->total_objects - 1].object.vertex == NULL) )
                scene->total_objects--;
}

///////////////////////////////////////////////////////////////////////////////

void obj_loader_init(SCENE *scene, int max_objects)
{
        scene->total_objects = 0;
        scene->obut = (OBJECTLINK *)malloc(max_objects * sizeof(OBJECTLINK));
}

///////////////////////////////////////////////////////////////////////////////

void free_object_structs(SCENE *scene)
{
        if (scene->obut != NULL) free(scene->obut);
}

///////////////////////////////////////////////////////////////////////////////

void all_objects_loaded(SCENE *scene)
{
        int i;

        scene->vertices_in_scene = 0;
        scene->faces_in_scene = 0;

        for (i = 0; i < scene->total_objects; i++)
        {
                scene->vertices_in_scene += scene->obut[i].object.vertices;
                scene->faces_in_scene += scene->obut[i].object.faces;
        }

        scene->rot = (VERTEX *)malloc(scene->vertices_in_scene * sizeof(VERTEX));
        scene->nor = (VERTEX *)malloc(scene->vertices_in_scene * sizeof(VERTEX));
}

///////////////////////////////////////////////////////////////////////////////

void calc_vertex_normals(SCENE *scene, unsigned int nro)
{
        int i, a, b;
        float len, ox, oy, oz, cx, cy, cz, cn;

        for (i = 0; i < scene->obut[nro].object.vertices; i++)
        {
                cx = cy = cz = cn = 0;

                for (a = 0; a < scene->obut[nro].object.faces; a++)
                        if ( (scene->obut[nro].object.face[a].a == i) ||  
                                (scene->obut[nro].object.face[a].b == i) ||  
                                (scene->obut[nro].object.face[a].c == i) )
                        {
                                calcnormal(scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].a].x,
                                           scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].a].y,
                                           scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].a].z,
                                           scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].b].x,
                                           scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].b].y,
                                           scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].b].z,
                                           scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].c].x,
                                           scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].c].y,
                                           scene->obut[nro].object.vertex[scene->obut[nro].object.face[a].c].z,
                                           &ox, &oy, &oz); 

                                len=sqrt(ox * ox + oy * oy + oz * oz);
                                if (len == 0.0) len = -1.0;

                                cx += ox / len; 
                                cy += oy / len; 
                                cz += oz / len;
                                cn += 1.0;
                        }

                if (cn > 0)
                {
                        len = sqrt(cx * cx + cy * cy + cz * cz); 
                        if (len == 0.0) len = -1.0;
                        scene->obut[nro].object.normal[i].x = (cx / len);
                        scene->obut[nro].object.normal[i].y = (cy / len);
                        scene->obut[nro].object.normal[i].z = (cz / len);

                        scene->obut[nro].object.uv[i].u = (cx / len);
                        scene->obut[nro].object.uv[i].v = (cy / len);
                }
                else
                {
                        scene->obut[nro].object.normal[i].x = 0;
                        scene->obut[nro].object.normal[i].y = 0;
                        scene->obut[nro].object.normal[i].z = 0;

                        scene->obut[nro].object.uv[i].u = 0;
                        scene->obut[nro].object.uv[i].v = 0;
                }

        }
}

///////////////////////////////////////////////////////////////////////////////

void calc_object_data(SCENE *scene, unsigned int nro)
{
        int a, faceA, faceB, faceC;
        float len, ox, oy, oz;

        for (a = 0; a < scene->obut[nro].object.vertices; a++)
        {
                scene->obut[nro].object.normal[a].x = 0;
                scene->obut[nro].object.normal[a].y = 0;
                scene->obut[nro].object.normal[a].z = 0;
                scene->obut[nro].object.uv[a].u = 0;
                scene->obut[nro].object.uv[a].v = 0;
        }

        for (a = 0; a < scene->obut[nro].object.faces; a++)
        {
                faceA = scene->obut[nro].object.face[a].a; // vertex 1 of face
                faceB = scene->obut[nro].object.face[a].b; // vertex 2 of face
                faceC = scene->obut[nro].object.face[a].c; // vertex 3 of face

                calcnormal(scene->obut[nro].object.vertex[faceA].x,
                           scene->obut[nro].object.vertex[faceA].y,
                           scene->obut[nro].object.vertex[faceA].z,
                           scene->obut[nro].object.vertex[faceB].x,
                           scene->obut[nro].object.vertex[faceB].y,
                           scene->obut[nro].object.vertex[faceB].z,
                           scene->obut[nro].object.vertex[faceC].x,
                           scene->obut[nro].object.vertex[faceC].y,
                           scene->obut[nro].object.vertex[faceC].z,
                           &ox, &oy, &oz);

                len=sqrt(ox * ox + oy * oy + oz * oz);
                if (len == 0.0) len = -1.0;
                
                ox /= len;
                oy /= len;
                oz /= len;

                scene->obut[nro].object.face[a].facenormal.x = ox;
                scene->obut[nro].object.face[a].facenormal.y = oy;
                scene->obut[nro].object.face[a].facenormal.z = oz;

                scene->obut[nro].object.normal[faceA].x += ox;
                scene->obut[nro].object.normal[faceA].y += oy;
                scene->obut[nro].object.normal[faceA].z += oz;

                scene->obut[nro].object.normal[faceB].x += ox;
                scene->obut[nro].object.normal[faceB].y += oy;
                scene->obut[nro].object.normal[faceB].z += oz;

                scene->obut[nro].object.normal[faceC].x += ox;
                scene->obut[nro].object.normal[faceC].y += oy;
                scene->obut[nro].object.normal[faceC].z += oz;
        }

        for (a = 0; a < scene->obut[nro].object.vertices; a++)
        {
                len = sqrt(pow(scene->obut[nro].object.normal[a].x,2) +
                           pow(scene->obut[nro].object.normal[a].y,2) +
                           pow(scene->obut[nro].object.normal[a].z,2));

                if (len == 0.0) len = -1.0;

                scene->obut[nro].object.normal[a].x /= len;
                scene->obut[nro].object.normal[a].y /= len;
                scene->obut[nro].object.normal[a].z /= len;

                scene->obut[nro].object.uv[a].u = scene->obut[nro].object.normal[a].x;
                scene->obut[nro].object.uv[a].v = scene->obut[nro].object.normal[a].y;
        }
}

///////////////////////////////////////////////////////////////////////////////

void set_flags(OBJECT *obu, unsigned int flgs, unsigned char textnro,
               unsigned char r, unsigned char g, unsigned char b)
{
        int a;

        switch (flgs & 15)
        {
                case p_gouraud:
                case p_flat:
                case p_phong:
                        flgs += (r << 8) + (g << 16) + (b << 24); // flat or gouraud, save rgb
                        break;

                case p_texture:
                case p_texturegouraud:
                case p_envmap:
                        flgs += ( (int)textnro << 8 ); // texture, save texturenro
                        break;

                default:
                        break;
        }

        for (a = 0; a < obu->faces; a++)
        {
                obu->face[a].flags = flgs;
        }
}

///////////////////////////////////////////////////////////////////////////////

signed int new_obj(SCENE *scene, char *filename,float mul,
                   float xpos, float ypos, float zpos,
                   unsigned int flags, unsigned char textnro,
                   unsigned char r, unsigned char g, unsigned char b)
{
        if (load_asc(filename, &scene->obut[scene->total_objects].object, mul) != 0)
        {
                dispose_object(scene, scene->total_objects);
                return -1;
        }

        calc_object_data(scene, scene->total_objects);
        set_flags(&scene->obut[scene->total_objects].object, flags, textnro, r, g, b);

        scene->obut[scene->total_objects].xpos = xpos;
        scene->obut[scene->total_objects].ypos = ypos;
        scene->obut[scene->total_objects].zpos = zpos;

        matrix_normalize(scene->obut[scene->total_objects].obj_matrix);

        scene->total_objects++;

        return (scene->total_objects - 1);      // handle
} 

///////////////////////////////////////////////////////////////////////////////

void object_update(SCENE *scene, unsigned int nro,
                   float xa, float ya, float za,
                   float s_rt, float s_dn, float s_fw)
{
        MATRIX chg;

        // prepare rotation matrix
        matrix_update(chg, sin(xa), sin(ya), sin(za),
                           cos(xa), cos(ya), cos(za));

        matrix_mul(scene->obut[nro].obj_matrix, chg); // rotate object

        // move object:
        scene->obut[nro].xpos += s_rt * scene->obut[nro].obj_matrix[0][0]+
                                 s_dn * scene->obut[nro].obj_matrix[0][1]+
                                 s_fw * scene->obut[nro].obj_matrix[0][2];

        scene->obut[nro].ypos += s_rt * scene->obut[nro].obj_matrix[1][0]+
                                 s_dn * scene->obut[nro].obj_matrix[1][1]+
                                 s_fw * scene->obut[nro].obj_matrix[1][2];

        scene->obut[nro].zpos += s_rt * scene->obut[nro].obj_matrix[2][0]+
                                 s_dn * scene->obut[nro].obj_matrix[2][1]+
                                 s_fw * scene->obut[nro].obj_matrix[2][2];
}

///////////////////////////////////////////////////////////////////////////////

void putobject(SCENE *scene, unsigned int nro, CAMERA *camera)
{
        MATRIX emat, nmat;

        int i;
        float zed;

        matrix_normalize(emat);          // prepare eye space matrix
        matrix_normalize(nmat);          // prepare matrix for normals

 
        matrix_mul(emat, scene->obut[nro].obj_matrix);

        matrix_mul(nmat, emat);
        matrix_mul(emat, camera->cam);    // camera

        scene->obut[nro].origin.x = scene->obut[nro].xpos - camera->xpos;
        scene->obut[nro].origin.y = scene->obut[nro].ypos - camera->ypos;
        scene->obut[nro].origin.z = scene->obut[nro].zpos - camera->zpos;

 
        matrix_mul_vector(scene->obut[nro].origin, &scene->obut[nro].origin, camera->cam);

        scene->obut[nro].hidden = 0;

        for (i = 0; i < scene->obut[nro].object.vertices; i++)
        {
   
                matrix_mul_vector(scene->obut[nro].object.vertex[i], &scene->rot[i + scene->total_rot_vertices], emat);

                scene->rot[i + scene->total_rot_vertices].x += scene->obut[nro].origin.x;
                scene->rot[i + scene->total_rot_vertices].y += scene->obut[nro].origin.y;
                scene->rot[i + scene->total_rot_vertices].z += scene->obut[nro].origin.z;

                matrix_mul_vector(scene->obut[nro].object.normal[i], &scene->nor[i + scene->total_rot_vertices], nmat);

                zed = scene->rot[i + scene->total_rot_vertices].z;
                if (zed == 0.0) zed = -1.0;

                zed = camera->perspective / zed;

                scene->rot[i + scene->total_rot_vertices].x *= zed;
                scene->rot[i + scene->total_rot_vertices].x += camera->center_x;
                scene->rot[i + scene->total_rot_vertices].x *= camera->x_suhde;

                scene->rot[i + scene->total_rot_vertices].y *= zed;
                scene->rot[i + scene->total_rot_vertices].y += camera->center_y;
                scene->rot[i + scene->total_rot_vertices].y *= camera->y_suhde;

        }

        scene->obut[nro].vertex_index = scene->total_rot_vertices;
        scene->total_rot_vertices += scene->obut[nro].object.vertices;
}


